/*
 * Soubor:  hash.c
 * Datum:   12.12.2008
 * Autor:   Barto Luk     xbarto55@fit.vutbr.cz                           *
 *          ervienka Juraj  xcervi09@fit.vutbr.cz                           *
 *          uri Michal     xduris01@fit.vutbr.cz                           *
 *          Elis Martin      xelism00@fit.vutbr.cz                           *
 *          Herich Martin    xheric00@fit.vutbr.cz                           *
 * Projekt: Implementace interpretu imperativnho jazyka IFJ08.              *
 * Popis:   Hashovac tabulka
 */



#include "hash.h"
#include "scanner.h"


/*
 * Funkce,kter vytvo a inicializuje hashovac tabulku
 */
thtab *htab_init()
{
  thtab *t=(thtab *)malloc(sizeof(thtab));            // alokace struktury tabulky
  if (t==NULL)
  {
    return NULL;
  }
  t->size=HASHTAB_SIZE;
  t->pole=malloc((t->size)*sizeof(tprvek *));              // alokace vlastn tabulky
  if (t==NULL)
  {
    return NULL;
  }

  for (int i = 0; i < (t->size); i++)                             // inicializace
    t->pole[i]=NULL;

  return t;
}

/*
 * Funkce pro zruen cel tabulky
 */
void htab_dispose (thtab *t)
{
  htab_delete(t);
  free(t->pole);
  free(t);
}


/*
 * Hashovacia funkcia - priradi danmu etzci vlastni hash
 */
unsigned int hash_function(const char *pomocna, unsigned htable_size)
{
  unsigned int a=0;
  unsigned char *p;

  for (p=(unsigned char*)pomocna; *p!='\0'; p++)
    a = 31*a + *p;

  return a % htable_size;
}

/*
 * Funkce pro vygenerovn nzvu konstanty
 */
void const_name(char *string, int number)
{
  memset(string,'0',CONSTLENGHT);      // vynulovn

  char *znak1=string,*znak2=string;
  char c;

  while (number>0)                         // pevod integeru na etzec
  {
    int i = 0;
    i = number%10;
    *znak1++=i+'0';
    number=number/10;
  }
  znak1=string+CONSTLENGHT-2;            // konec etzce
  *(string+CONSTLENGHT-1)='\0';       // pidn koncovho znaku
  *znak1='$';

  for (int i=0; i<CONSTLENGHT/2; i++)              // reverzace retzce s nzvem
  {
    c=*znak1;
    *znak1=*znak2;
    *znak2=c;
    znak1--;
    znak2++;
  }
}


/*
 * Funkce kter vyhledv a vkld data do tabulky
 */
tprvek *htable_lookup (thtab *t, char *jmeno, char *const_value, int typ)
{
  if (jmeno==NULL)                                    // je to konstanta?
  {
    static int i=0;                                  // potadlo konstant - aby bylo zajitno vytvoen jedinenho nzvu

    jmeno=malloc(CONSTLENGHT*sizeof(char));
    if (jmeno==NULL)
    {
      return NULL;
    }
    const_name(jmeno,i++);
  }

  unsigned int index=hash_function (jmeno, t->size);        // zskn hashe, pomoc funkce hash_funkcion
  tprvek *tmp=t->pole[index];
  tprvek *tmp_prev=tmp;
  int first=0;

  while (tmp!=NULL)                                        // vyhledn poloky
  {
    if (strcmp(tmp->jmeno, jmeno)==0)
    {
      free(jmeno);
      return tmp;                                          // hledan zznam u v tabulce je
    }
    tmp_prev=tmp;
    tmp=tmp->dalsi;
    first++;
  }

  tprvek *novyPrvek=(tprvek *)malloc(sizeof(tprvek));     // vloen novho zznamu
  if (novyPrvek==NULL) return NULL;

  novyPrvek->jmeno=jmeno;                                            // samotn vkldn jednotlivch dat

  switch (typ)
  {
    case (IDENTIFIER): novyPrvek->typ=UNDEFINED;
                          break;

    case (RESW): novyPrvek->typ=RESW;
                    break;

    case (N_INT): novyPrvek->typ=N_INT;
                    if (const_value!=NULL)
                    {
                      novyPrvek->hodnota.int_cislo=atoi(const_value);
                      free(const_value);
                    }
                    break;

    case (N_DOUBLE): novyPrvek->typ=N_DOUBLE;
                 if (const_value!=NULL)
                 {
                   novyPrvek->hodnota.double_cislo=atof(const_value);
                   free(const_value);
                 }
                 break;

    case (RETKONS): novyPrvek->typ=RETKONS;
                    novyPrvek->hodnota.retezec=const_value;
                    break;

    default: break;
  }

  novyPrvek->dalsi=NULL;

  if (first==0) t->pole[index]=novyPrvek;              // vkladme priamo do tabuky, nie na koniec zoznamu
  else tmp_prev->dalsi=novyPrvek;

  return novyPrvek;
}

/*
 * Funkce ru veker poloky, kter byly dosud v tabulce
 */
void htab_delete (thtab *t)
{
  tprvek *pomocna;

  for (int i=0; i<t->size; i++)
  {
    while (t->pole[i]!=NULL)
    {
      pomocna=t->pole[i];
      t->pole[i]=t->pole[i]->dalsi;
      free(pomocna->jmeno);
      if (pomocna->typ==RETKONS) free(pomocna->hodnota.retezec);
      free(pomocna);
    }
  }
}
