/*
 * Soubor:  scanner.c
 * Datum:   14.12.2008
 * Autori:  Barto Luk     xbarto55@fit.vutbr.cz                           *
 *          ervienka Juraj  xcervi09@fit.vutbr.cz                           *
 *          uri Michal     xduris01@fit.vutbr.cz                           *
 *          Elis Martin      xelism00@fit.vutbr.cz                           *
 *          Herich Martin    xheric00@fit.vutbr.cz                           *
 * Projekt: Implementace interpretu imperativnho jazyka IFJ08.              *
 * Popis:   Lexikln analyztor
 */

#include "scanner.h"

//deklarace konstant

#define LONG_MAX        2147483647L
#define LONG_MIN        (-LONG_MAX-1)
#define INT_MAX	        2147483647
#define INT_MIN	        (-INT_MAX-1)

//deklarace funkci

int KonecRadku (int x);
int Nactiznak(FILE *soubor, unsigned int *sloupec, unsigned int *radek);
void Odeberznak (int znak, FILE *soubor, unsigned int *sloupec, unsigned int *radek);

const char *keywords[]= { "cin", "concat", "cout", "double", "else", "find", "if", "int", "length", "main", "return", "sort", "string", "while"};

/*
 * Mozne stavy konecneho automatu :
 */
typedef enum {
               START,
               ID,
               INTEGER,
               DOUBLE_E,
               DOUBLE_OK,
               DOUBLE_T,
               DOUBLE_S,
               DOUBLE_N,
               GREATER_OR_EQUAL,
               LESSER_OR_EQUAL,
               INSTATE,
               OUTSTATE,
               s1,
               s2,
               s3,
               s4,
               CARKA,
               LBRACE,
               RBRACE,
               LESSERTHEN,
               GREATERTHEN,
               POROVNANI,
               ROVNO,
               STREDNIK,
               KOMENTAR,
               KONEC,
               RETEZEC,
               MUJ_EOF,
               s10,
               s11,
               s12,
               ASIESC,
} Tstate;

static unsigned int pocetsloupcu, sloupec = 0, radek = 1;

///////////////////////////////////////////////////////////////////////////////
/*
 * Funkce nat znaky ze souboru a rozpoznv lexmy. Funkce vrac jejich typy.
 * Pokud jde o identifiktor, klov slovo nebo o konstantu
 * vlo je do hahsovac tabulky.
 * Funkce pracuje na principu deterministickho konenho automatu.
 */
int LoadToken (FILE *fr, thtab *t, tprvek **prvek)
{

int nova_alokace = TRUE;

  string str;
  Tstate state = START;
  int znak;
  do
  {
    znak=Nactiznak(fr, &sloupec, &radek);
    switch(state)
    {
      case START:                                                               //pocatecni stav
      {
        if (isspace(znak))                                                      //preskoceni prazdnych znaku
        {
          while (((isspace(znak)) || KonecRadku(znak))&& (znak!=EOF))
          {
            znak=Nactiznak(fr, &sloupec, &radek);
          }
          Odeberznak(znak, fr, &sloupec, &radek);
          state = START;
        }
        else
        {
          if(isalpha(znak))
          {
            if(nova_alokace == TRUE)
            {
              str = str_init();
              nova_alokace = FALSE;
            }
            state = ID;
            Odeberznak(znak, fr, &sloupec, &radek);
          }
          else if (isdigit(znak))
          {
            if(nova_alokace == TRUE)
            {
              str = str_init();
              nova_alokace = FALSE;
            }
            state=INTEGER;
            Odeberznak(znak, fr, &sloupec, &radek);
          }
          else if(znak == '/') state = s1;
          else if(znak == '*') state = s2;
          else if(znak == '+') state = s3;
          else if(znak == '-') state = s4;
          else if(znak == '<') state = LESSERTHEN;
          else if(znak == '>') state = GREATERTHEN;
          else if(znak == '=') state = ROVNO;
          else if(znak == ';') state = STREDNIK;
          else if(znak == '"') state = RETEZEC;
          else if(znak == '!') state = s10;
          else if(znak == ')') state = s11;
          else if(znak == '(') state = s12;
          else if(znak == '{') state = LBRACE;
          else if(znak == '}') state = RBRACE;
          else if(znak == ',') state = CARKA;
          else if(znak == EOF) state = MUJ_EOF;
          else
          {
            error = ERR_LEX;
            return BADLEXEM;
          }
        }
      }
      break;

      case ID:                                                                  //identifikator
      {
        if(isalpha(znak) || isdigit(znak))
        {
          add_char( &str, znak);
        }
        else
        {
          int i = 0;
          for (i = 0; i < 14; i++)
          {
            if(strcmp( str.text, keywords[i]) == 0)
            {
              Odeberznak(znak, fr, &sloupec, &radek);
              *prvek=htable_lookup (t, strGetStr(&str), NULL, RESW);            //vloen do hashovac tabulky
              return RESW;                                                      //identifikator je klicove slovo
            }
          }
          Odeberznak(znak, fr, &sloupec, &radek);
          *prvek=htable_lookup (t, strGetStr(&str), NULL, IDENTIFIER);          //vloen do hashovac tabulky
          return IDENTIFIER;
        }
      }
      break;

      case INTEGER:                                                             //integer
      {
        if(isdigit(znak))
        {
          add_char( &str, znak);
          state = INTEGER;
        }
        else if(znak == '.')
        {
          state = DOUBLE_T;
          add_char( &str, znak);
        }
        else if(znak == 'e' || znak == 'E')
        {
          state = DOUBLE_E;
          add_char( &str, znak);
        }
        else if(isalpha(znak))
        {
          add_char( &str, znak);
          state = ID;
        }
        else
        {
          Odeberznak(znak, fr, &sloupec, &radek);
          *prvek=htable_lookup (t, NULL, strGetStr(&str), N_INT);               //vloen do hashovac tabulky
          return N_INT;
        }
      }
      break;

      case DOUBLE_T:                                                            //kontrola spravne zapsaneho double
      {
        if(isdigit(znak))
        {
          add_char( &str, znak);
          state = DOUBLE_OK;
        }
        else
        {
          error = ERR_LEX;
          return BADLEXEM;
        }
      }
      break;

      case DOUBLE_OK:                                                           //double
      {
        if(isdigit(znak))
        {
          add_char( &str, znak);
          state = DOUBLE_OK;
        }
        else if(znak == 'e' || znak == 'E')
        {
          state = DOUBLE_E;
          add_char( &str, znak);
        }
        else
        {
          Odeberznak(znak, fr, &sloupec, &radek);
          *prvek=htable_lookup (t, NULL, strGetStr(&str), N_DOUBLE);            //vloen do hashovac tabulky
          return N_DOUBLE;
        }
      }
      break;

      case DOUBLE_E:                                                            //kontrola spravne zapsaneho double
      {
        if(znak == '+' || znak == '-')
        {
          add_char( &str, znak);
          state = DOUBLE_N;
        }
        else if(isdigit(znak))
        {
          add_char( &str, znak);
          state = DOUBLE_S;
        }
        else
        {
          error = ERR_LEX;
          return BADLEXEM;
        }
      }
      break;

      case DOUBLE_S:                                                            //double
      {
        if(isdigit(znak))
        {
          add_char( &str, znak);
          state = DOUBLE_S;
        }
        else
        {
          Odeberznak(znak, fr, &sloupec, &radek);
          *prvek=htable_lookup (t, NULL, strGetStr(&str), N_DOUBLE);            //vloen do hashovac tabulky
          return N_DOUBLE;
        }
      }
      break;

      case DOUBLE_N:                                                            //kontrola spravne zapsaneho double
      {
        if(isdigit(znak))
        {
          add_char( &str, znak);
          state = DOUBLE_S;
        }
        else
        {
          error = ERR_LEX;
          return BADLEXEM;
        }
      }
      break;

      case s1:
      {
        if(znak == '*')                                                         //komentar
        {
          state = KOMENTAR;
        }
        else
        {
          Odeberznak(znak, fr, &sloupec, &radek);
          return DIV;
        }
      }
      break;

      case KOMENTAR:                                                            //blokovy komentar
      {
        if(znak == '*')
        {
          state = KONEC;
        }
        else if (znak==EOF)
        {
          error=ERR_LEX;
          return BADLEXEM;
        }
        else
        {
          state = KOMENTAR;
        }
      }
      break;

      case KONEC:                                                               //mozny konec blokoveho komentare
      {
        if(znak == '/')
        {
          state = START;
        }
        else if (znak==EOF)
        {
          error=ERR_LEX;
          return BADLEXEM;
        }
        else
        {
          state = KOMENTAR;
        }
      }
      break;

      case s2:
      {
        Odeberznak(znak, fr, &sloupec, &radek);
        return MUL;
      }
      break;

      case s3:
      {
        Odeberznak(znak, fr, &sloupec, &radek);
        return PLUS;
      }
      break;

      case s4:
      {
        Odeberznak(znak, fr, &sloupec, &radek);
        return MINUS;
      }
      break;

      case LESSERTHEN:
      {
        if(znak == '=')
        {
          state = LESSER_OR_EQUAL;
        }
        else if(znak == '<')
        {
          state = OUTSTATE;
        }
        else
        {
          Odeberznak(znak, fr, &sloupec, &radek);
          return LESS;
        }
      }
      break;

      case LESSER_OR_EQUAL:
      {
        Odeberznak(znak, fr, &sloupec, &radek);
        return LESSOREQUAL;
      }
      break;

      case CARKA:
      {
        Odeberznak(znak, fr, &sloupec, &radek);
        return COMMA;
      }
      break;

      case GREATER_OR_EQUAL:
      {
        Odeberznak(znak, fr, &sloupec, &radek);
        return GREATEROREQUAL;
      }
      break;

      case GREATERTHEN:
      {
        if(znak == '=')
        {
          state = GREATER_OR_EQUAL;
        }
        else if(znak == '>')
        {
          state = INSTATE;
        }
        else
        {
          Odeberznak(znak, fr, &sloupec, &radek);
          return GREATER;
        }
      }
      break;

      case INSTATE:
      {
        Odeberznak(znak, fr, &sloupec, &radek);
        return IN;
      }
      break;

      case OUTSTATE:
      {
        Odeberznak(znak, fr, &sloupec, &radek);
        return OUT;
      }
      break;

      case ROVNO:
      {
        if(znak == '=')
        {
          state = POROVNANI;
        }
        else
        {
          Odeberznak(znak, fr, &sloupec, &radek);
          return EQUAL;
        }
      }
      break;

      case POROVNANI:
      {
        Odeberznak(znak, fr, &sloupec, &radek);
        return ISEQUAL;
      }
      break;

      case STREDNIK:
      {
        Odeberznak(znak, fr, &sloupec, &radek);
        return SEMICOLON;
      }
      break;

      case RETEZEC:                                                             //retezec
      {
        if(nova_alokace == TRUE)
        {
          str = str_init();
          nova_alokace = FALSE;
        }
        if(znak == '"')
        {
          *prvek=htable_lookup (t, NULL, strGetStr(&str), RETKONS);             //vloen do hashovac tabulky
          return RETKONS;
        }
        else if (znak==EOF)
        {
          error=ERR_LEX;
          str_free(&str);
          return BADLEXEM;
        }
        else
        {
          if(znak == '\\')                                                      //escape sekvence
          {
            state = ASIESC;
          }
          else
          {
            add_char( &str, znak);
          }
        }
      }
      break;

      case s10:
      {
        if(znak == '=')
        {
          return NOT_EQUAL;
        }
        else
        {
          error = ERR_LEX;
          return BADLEXEM;
        }
      }
      break;

      case LBRACE:
      {
        Odeberznak(znak, fr, &sloupec, &radek);
        return L_BRACE;
      }
      break;

      case RBRACE:
      {
        Odeberznak(znak, fr, &sloupec, &radek);
        return R_BRACE;
      }
      break;

      case s11:
      {
        Odeberznak(znak, fr, &sloupec, &radek);
        return R_PARENTHESIS;
      }
      break;

      case s12:
      {
        Odeberznak(znak, fr, &sloupec, &radek);
        return L_PARENTHESIS;
      }
      break;

      case ASIESC:                                                              //escape sekvence
      {
        state = RETEZEC;
        if (znak==EOF)
        {
          error=ERR_LEX;
          str_free(&str);
          return BADLEXEM;
        }
        switch(znak)
        {
          case 'n':
          {
            add_char( &str, 'n');
            break;
          }

          case '\\':
          {
            add_char( &str, '\\');
            break;
          }

          case 't':
          {
            add_char( &str, 't');
            break;
          }

          case '"':
          {
            add_char( &str, '"');
            break;
          }

          default:
          {
            add_char( &str, znak);
            break;
          }
        }
      }
      break;

      case MUJ_EOF:                                                             //konecny stav, nalezen konec souboru
      {
        return MUJEOF;
      }
      break;
    }
  }while(1);
}

/*
 * Funkce pro nalezeni konce radku.
 */

int KonecRadku (int x) {
  if(x == '\n')
  {
    return TRUE;
  }
  else
  {
    return FALSE;
  }
}

/*
 * Funkce pro prochazeni souboru a nacitani znaku po radcich.
 */
int Nactiznak(FILE *soubor, unsigned int *sloupec, unsigned int *radek) {

  int p;
  p = fgetc(soubor);

  if(KonecRadku(p) == TRUE)
  {
    (*sloupec) = 0;
    (*radek)++;
    pocetsloupcu = (*sloupec);
  }
  else
  {
    (*sloupec)++;
  }
  return (p);
}

/*
 * Funkce vraci nacteny znak zpet do souboru.
 */
void Odeberznak (int znak, FILE *soubor, unsigned int *sloupec, unsigned int *radek)
{
 ungetc (znak,soubor);
 if (znak == '\n')
 {
    (*radek)--;
    (*sloupec) = pocetsloupcu;
 }
 else (*sloupec)--;
}
